/* io_lib.h	  */
/* (c) by Marek Zelem */

#ifndef _IO_LIB_H
#define _IO_LIB_H

#include <mmm/dynamic.h>

#define	NOTNULL		( (void *) 1 )

#define	IO_RD	0
#define	IO_WR	1
#define	IO_EX	2
#define	IO_TM	3
#define	IO_TM_S	4		/* ked namiesto tych TM ktore nestiha */
#define	IO_CLOSED	7	/* rd , wr , ex  was closed */
#define	IO_INIT	8
#define	IO_DEL	9
#define IO_RDERR	10	/* generated by ioio.c if read return error */
#define IO_WRERR	11	/* gen. by ioio.c if dfifo_read return error */
#define IO_WREXTR	12	/* gen. by ioio.c if outbuf=NOTNULL  */
#define IO_LAST_SYSTEM	15	/* 0 -15 rezervovane pre system */

#define	IO_HELLO	16	/* for free use */

typedef unsigned long	ioid_t;
#define	IO_NULLID	(0xffffffff)

struct evlist {
		struct evlist	*next;
		long		evid;	/* event identifacator */
		ioid_t		caller;
		unsigned long	dtime;	/* in miliseconds */
		int		what;
		void		*arg;
		void(*free)(void *);
	      };
typedef struct evlist evlist_t;

struct io {
	struct io	*head;
	struct io	*next;
	struct io	*prev;
	evlist_t	*evlist;	/* linked list of events */
	ioid_t	id;	/* object identification */
	int	rd;	/* rd,wr,ex - close on exit(this entry) */
	int	wr;
	int	ex;
	int	prior;			/* priority */
	unsigned long	msec,nmsec;	/* timer */
	int	user1;			/* for free use */
	int	user2;			/* for free use */
	int	(*func)(struct io *io,int what,void *arg);	/* function */
	int	(*ufn)(struct io *io,int what,void *arg);   /* for free use */
	void	*ptr;		/* ! in first is reserved for iot_t */
	void	*ptr2;		/* for extend structure */
	void	*inbuf;		/* if NULL then not ready for read */
	void	*outbuf;	/* if NULL then not ready for write */
	ioid_t	caller;		/* id of caller */
	ioid_t	user_id;	/* for free use */
	  };
typedef struct io io_t;

#define proto_ia(prior,msec,nmsec,func,ufn,user1,user2,prt,ptr2,inbuf,outbuf) \
		{ NULL,NULL,NULL,NULL,0,-1,-1,-1,(prior),(msec),(nmsec),\
		  (user1),(user2),(func),(ufn),(ptr),(ptr2),(inbuf),(outbuf), \
		  0xffffffff,0xffffffff \
		}

#define proto_io(prior,func,ufn,ptr,ptr2)	\
		{ NULL,NULL,NULL,NULL,0,-1,-1,-1,(prior),0,0,\
		  0,0,(func),(ufn),(ptr),(ptr2),NULL,NULL, \
		  0xffffffff,0xffffffff \
		}

int io_verlib( char *buf );

/* io */

io_t *io_init( int(*func)(io_t*,int,void*) );
int io_final( io_t *io );
int io_closeall( io_t *io );	/* close all filedes - krizove riesenie */
io_t *io_first( io_t *io );
io_t *io_find( io_t *io, long id );
ioid_t io_id( io_t *io );
int io_exist( io_t *io, io_t *f );

int io_setopen(io_t *io, int fd);
int io_setclose(io_t *io, int fd);
int io_close(io_t *io, int fd);

io_t *io_add(io_t *io, io_t *prototyp );
int io_delio( io_t *io );
int io_chprior( io_t *io, int prior );
int io_chrd( io_t *io, int rd );
int io_chwr( io_t *io, int wr );
int io_chex( io_t *io, int ex );
int io_chtimer( io_t *io, unsigned long msec, unsigned long nmsec );

int io_exit(io_t *io, int retval);	/* retval!=0 */
int io_callio( io_t *io, int what, void *arg );
/* io_call - NEPOUZIVAT  , ak what=IO_DEL potom to zrusi objekt */
int io_call( io_t *io, ioid_t id, int what, void *arg );
/* hodnoty pre argument ffree funkcie io_schedule */
#define	STRING	((void*)2)
#define	INT	((void*)3)
#define	DFIFO	((void(*)(void*))dfifo_delete)
/* hodnoty pre argument evid funkcie io_schedule */
#define AT_END	1
long io_schedule( io_t *io, ioid_t id, long evid, unsigned long dtime, int what, void *arg, void(*ffree)(void*) );
#define	io_sched_me(io,dtime,what,arg,ffree)	\
		io_schedule((io),IO_NULLID,0,(dtime),(what),(arg),(ffree))
int io_do( io_t *io );

/*	io_extend	*/

#define	is_ioext_on(io)	(((io)->ptr2)!=NULL)
#define	ioext(io)	(((ioextend_t*)((io)->ptr2))+1)
#define	ioextn(io)	((int)(((ioextend_t*)((io)->ptr2))->num))
#define	ioextptr(io)	(((ioextend_t*)((io)->ptr2))->ptr)

typedef struct	{
			short	magic;
			short	user;
			long	num;
			long	num1;
			void	*ptr;
		} ioextend_t;

/* io_extend use ptr2 of io */
int io_extend_init( io_t *io, int n );
int io_extend_resize( io_t *io, int n );
int io_extend_del( io_t *io );

ioextend_t *io_extend_alloc( io_t *io, short magic );
ioextend_t *io_extend_find( io_t *io, short magic );
int io_extend_release( io_t *io, short magic );


/* iouf_all_to_all
 * write to all who use this function
 */
int iouf_all_to_all( io_t *io, int what, void *arg );

int io_write( io_t *io, char *buf, int len );
int io_bufwrite( io_t *io, dfifo_t *buf );
int io_printf( io_t *io, char *fmt, ... );
int io_writebuf_erase( io_t *io );	/* return -1 if rd < 0 */


/* load and unload files */

char *load_file( char *filename, long *len );
int unload_file( char *mem, long len );

/* ----- io function ----- */

/* iof_in_out
 * pre RD: cita z rd a zapisuje do inbuf a vola ufn()
 * pre WR: cita z outbuf a zapisuje do wr, ak je prazdny outbuf, tak vola ufn()
 * in/out buffer je typu dfifo_t
 * na vsetky ostatne akcie vola ufn()
 * ak rd,wr a ex zatvorene, tak generuje akciu IO_CLOSED
 * ak sa zapis do wr nepodari, tak sa objekt ukonci
 */
int iof_in_out( io_t *io, int what, void *arg );

/* iof_input_line
 * pre RD: cita z rd a zapisuje do inbuf, ak \n tak vola ufn()
 * inak ako pre iof_in_out
 */
int iof_input_line( io_t *io, int what, void *arg );

/* iof_input_char
 * pre RD: cita z rd a zapisuje do inbuf, za kazdym znakom vola ufn()
 * inak ako pre iof_in_out
 */
int iof_input_char( io_t *io, int what, void *arg );

/*  iof_charfunc	*/

#define	CHARFUNC_EVER		0x100
#define	CHARFUNC_DEFAULT	0x101
#define	CHARFUNC_AFTER		0x102

typedef struct	{
			int	c;
			int(*func)(io_t *io, char c);
		}  charfunc_t;

/* iof_charfunc
 * pre RD: cita z rd a zapisuje do inbuf, za kazdym znakom prelieza pole
 *         charfunc a spusta prislusnu funkciu.
 * inak ako pre iof_in_out
 * io->ptr is charfunc_t *
 */
int iof_charfunc( io_t *io, int what, void *arg );

#endif _IO_LIB_H

