/* Indexed Memory in file
 * (c)31.10.1999 by Marek Zelem
 */

#include <stdlib.h>
#include <mmm/imemory.h>
#include <mmm/mutex.h>

int imem_errno=0;

static int get_access( struct imem_s *m, int id )
{ int t=0;
  struct fmem_head_s *f;
	if( (f=m->fmem->file_head(m->fmem,id>>24))==NULL )
		return(-1);
	mutex_wait(&(f->mutex[1]))
	{	if( (t++)>=100 )
		{	imem_errno=IME_MUTEX;
			return(0);
		}
		sched_yield();
	}
	imem_errno=0;
	return(1);
}
static void put_access( struct imem_s *m, int id )
{ struct fmem_head_s *f;
	if( (f=m->fmem->file_head(m->fmem,id>>24))==NULL )
		return;
	mutex_signal(&(f->mutex[1]));
}

static void imem_destroy(struct imem_s *m)
{
	m->fmem->destroy(m->fmem);
}

static int imem_create_index(struct imem_s *m, int memid)
{ fmem_t *fm=m->fmem;
  struct fmem_head_s *fh;
  long addr;
  long *in;
  int i;
	if( !get_access(m,memid) )	return(-1);
  	if( (fh=fm->file_head(fm,memid))==NULL )
	{	put_access(m,memid);
		return(-1);
	}
	if( fm->get(fm,fh->file_magic_id)!=NULL )
	{	put_access(m,memid);
		return(-2);
	}
	if( (addr=fm->malloc(fm,256*sizeof(long),memid))==NOMEMID )
	{	put_access(m,memid);
		return(-3);
	}
	if( (addr&FMEMADDR_FILEID)!=(fh->file_magic_id&FMEMADDR_FILEID) )
	{	fm->free(fm,addr);
		put_access(m,memid);
		return(-4);
	}
	if( (in=fm->get(fm,addr))==NULL )
	{	put_access(m,memid);
		return(-3);
	}
	for(i=0;i<256;i++)
		in[i]=NOMEMID;
	fh->file_magic_id=addr;
	put_access(m,memid);
	return(0);
}

static long new_low_index( fmem_t *fm, long newid, long addr, int level )
{ long *in,n,*in2;
  struct fmem_blok_head_s *ih;
  int i,f;
	if( (in=fm->get(fm,addr))==NULL )
		return(NOMEMID);
	for(f=-1,i=0;i<256;i++)
	{	if(in[i]==NOMEMID)
		{	if( level==1 )
			{	in[i]=0;	/* obsadene, neplatna hodn. */
				return((newid&FMEMADDR_FILEID)|
					((newid<<8)&FMEMADDR_BLOKID)|i);
			}
			if( f<0 )	f=i;
			continue;
		}
		if( level==1 )	continue;
		if( (n=new_low_index(fm,(newid&FMEMADDR_FILEID)|
				((newid<<8)&FMEMADDR_BLOKID)|i,
				in[i],level-1))!=NOMEMID )
			return(n);
	}
	if( f<0 || level<=1 )	return(NOMEMID);
	if( (in[f]=fm->malloc(fm,256*sizeof(long),newid>>24))==NOMEMID )
		return(NOMEMID);
	if( (in2=fm->get(fm,in[f]))==NULL )
		return(NOMEMID);
	for(i=0;i<256;i++)
		in2[i]=NOMEMID;
	ih=fm->head(fm,in[f]);
	ih->user_adr=addr;
	ih->user_id=f;
	return( new_low_index(fm,(newid&FMEMADDR_FILEID)|
				((newid<<8)&FMEMADDR_BLOKID)|f,
				in[f],level-1) );
}

static long new_index( fmem_t *fm, int memid )
{ struct fmem_head_s *fh;
  	if( (fh=fm->file_head(fm,memid))==NULL )
		return(NOMEMID);
	return( new_low_index(fm,memid<<24,fh->file_magic_id,3) );
}

static long *get_index( fmem_t *fm, long id )
{ struct fmem_head_s *fh;
  long *in;
	if( (fh=fm->file_head(fm,id>>24))==NULL )
		return(NULL);
	if( (in=fm->get(fm,fh->file_magic_id))==NULL )
		return(NULL);
	if( (in=fm->get(fm,in[(id>>16)&0x00ff]))==NULL )
		return(NULL);
	if( (in=fm->get(fm,in[(id>>8)&0x00ff]))==NULL )
		return(NULL);
	return( &(in[(id)&0x00ff]) );
}

static long get_ind_id( fmem_t *fm, long id )
{ struct fmem_head_s *fh;
  long *in;
	if( (fh=fm->file_head(fm,id>>24))==NULL )
		return(0);
	if( (in=fm->get(fm,fh->file_magic_id))==NULL )
		return(0);
	if( (in=fm->get(fm,in[(id>>16)&0x00ff]))==NULL )
		return(0);
	return(in[(id>>8)&0x00ff]);
}

static long imem_new(struct imem_s *m, long size, int memid)
{ long ind,bid;
  struct fmem_blok_head_s *bh;
	if( !get_access(m,memid) )	return(-1);
	if( (ind=new_index(m->fmem,memid))==NOMEMID )
	{	put_access(m,memid);
		return(NOMEMID);
	}
	if( (bid=m->fmem->malloc(m->fmem,size,memid))==NOMEMID )
	{	*(get_index(m->fmem,ind))=NOMEMID;
		put_access(m,memid);
		return(NOMEMID);
	}
	*(get_index(m->fmem,ind))=bid;
	bh=m->fmem->head(m->fmem,bid);
	bh->user_id=ind&0x00ff;
	bh->user_adr=get_ind_id(m->fmem,ind);
	put_access(m,memid);
	return(ind);
}

static void *imem_get(struct imem_s *m, long id)
{ long *i;
  void *r;
	if( !get_access(m,id) )	return(NULL);
	if( (i=get_index(m->fmem,id))==NULL )
		r=NULL;
	else	r=m->fmem->get(m->fmem,*i);
	put_access(m,id);
	return( r );
}

static long imem_size(struct imem_s *m, long id)
{ long *i,r;
	if( !get_access(m,id) )	return(-1);
	if( (i=get_index(m->fmem,id))==NULL )
		r=0;
	else	r=m->fmem->get_size(m->fmem,*i);
	put_access(m,id);
	return( r );
}

static int imem_resize(struct imem_s *m, long id, long size)
{ long *i,n;
	if( !get_access(m,id) )	return(-1);
	if( (i=get_index(m->fmem,id))==NULL )
	{	put_access(m,id);
		return(-2);
	}
	n=m->fmem->realloc(m->fmem,*i,size);
	if( n==NOMEMID )
	{	put_access(m,id);
		return(-1);
	}
	*i=n;
	put_access(m,id);
	return(0);
}

static int imem_inc_use(struct imem_s *m, long id)
{ long *i;
  struct fmem_blok_head_s *bh;
	if( !get_access(m,id) )	return(-1);
	if( (i=get_index(m->fmem,id))==NULL )
	{	put_access(m,id);
		return(-1);
	}
	if( (bh=m->fmem->head(m->fmem,*i))==NULL )
	{	put_access(m,id);
		return(-1);
	}
	bh->usecount++;
	put_access(m,id);
	return(bh->usecount);
}

static int imem_dec_use(struct imem_s *m, long id)
{ long *i;
  struct fmem_blok_head_s *bh;
	if( !get_access(m,id) )	return(-1);
	if( (i=get_index(m->fmem,id))==NULL )
	{	put_access(m,id);
		return(-1);
	}
	if( (bh=m->fmem->head(m->fmem,*i))==NULL )
	{	put_access(m,id);
		return(-1);
	}
	if( bh->usecount<=1 )
	{	m->fmem->free(m->fmem,*i);
		*i=NOMEMID;
		put_access(m,id);
		return(0);
	}
	bh->usecount--;
	put_access(m,id);
	return(bh->usecount);
}

imem_t *imem_create( fmem_t *fmem )
{ imem_t *i;
	if( fmem==NULL )
		return(NULL);
	if( (i=malloc(sizeof(imem_t)))==NULL )
		return(NULL);
	i->destroy=imem_destroy;
	i->create_index=imem_create_index;
	i->new=imem_new;
	i->get=imem_get;
	i->size=imem_size;
	i->resize=imem_resize;
	i->inc_use=imem_inc_use;
	i->dec_use=imem_dec_use;
	i->fmem=fmem;
	return(i);
}

