/* User Interface - types
 * (c) by Marek Zelem
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <mmm/menu.h>

int declcmp( const char *a, const char *b )
{
	return(strcmp(a,b));	/* prerobit */
}

static void th_std_free( long d )
{
	if( d!=0 )	free((char*)d);
}

static long th_std_dup( long d )
{	return(d);
}

static long th_char_input( int fd )
{ char c;
  	if( read(fd,&c,1)==1 )
		return((long)c);
	return(0);
}

char *read_string( int fd )
{ char *p,*pp;
  int len,size;
  	len=0;
	size=16;
	if( (p=malloc(size))==NULL )	return(NULL);
  	while( read(fd,p+len,1)==1 )
	{	if( p[len]==0 || p[len]=='\n' )
			break;
		len++;
		if( len>=size )
		{	size=len+16;
			if( (p=realloc((pp=p),size))==NULL )
			{	pp[len-1]=0;
				return(pp);
			}
		}
	}
	p[len]=0;
	return(p);
}

static long th_char_p_input( int fd )
{ char *c;
  	if( (c=read_string(fd))!=NULL )
		return((long)c);
	return(0);
}

static long th_char_p_dup( long d )
{
	if( d==0 )	return(0);
	return( (long)strdup((char*)d) );
}

static int conv_zn( char c )
{
	if( c>='0' && c<='9' )
		return(c-'0');
	if( c>='a' && c<='f' )
		return(c-'a'+10);
	return(1000);
}

static long read_long( int fd )
{ long d=0,i;
  int base=10,sign=1,z=0;
  char c;
  	while( read(fd,&c,1)==1 )
	{	if( z==0 && c=='-' )	sign=-1;
		else if( z==0 && c=='0' )
		{	base=8;	z=1;	}
		else if( (z==0 || z==1) && c=='x' )
		{	base=16; z=2;	}
		else if( (i=(long)conv_zn(tolower(c)))<base )
		{	d*=base;
			d+=i;
			z=2;
		}
		else	return(d*sign);
	}
	return(d*sign);
}

#define TH_type_INPUT(type)	\
{ type c;\
	c=(type)(read_long(fd));\
	return((long)c);\
}
static long th_u_short_input( int fd )		TH_type_INPUT(unsigned short)
static long th_s_short_input( int fd )		TH_type_INPUT(signed short)
static long th_u_int_input( int fd )		TH_type_INPUT(unsigned int)
static long th_s_int_input( int fd )		TH_type_INPUT(signed int)
static long th_u_long_input( int fd )		TH_type_INPUT(unsigned long)
static long th_s_long_input( int fd )		TH_type_INPUT(signed long)

#define TH_type_PRINT(pr,type,isend)	\
{ char buf[64];\
	sprintf(buf,"%"pr,(type)d);\
	write(fd,buf,strlen(buf));\
	return( sizeof(type)*((isend)?-1:1) );\
}
static long th_char_print( int fd, long d )	TH_type_PRINT("c",char,d==0)
static long th_char_p_print( int fd, long d )	TH_type_PRINT("s",char *,d==0)
static long th_u_short_print( int fd, long d )	TH_type_PRINT("u",unsigned short,0)
static long th_s_short_print( int fd, long d )	TH_type_PRINT("d",signed short,0)
static long th_u_int_print( int fd, long d )	TH_type_PRINT("u",unsigned int,0)
static long th_s_int_print( int fd, long d )	TH_type_PRINT("d",signed int,0)
static long th_u_long_print( int fd, long d )	TH_type_PRINT("lu",unsigned long,0)
static long th_s_long_print( int fd, long d )	TH_type_PRINT("ld",signed long,0)

#define TH_type_STORE(type,isend)	\
{\
	if( addr!=NULL )\
		*((type *)addr)=(type)d;\
	return( sizeof(type)*((isend)?-1:1) );\
}
static long th_char_store( void *addr, long d )	TH_type_STORE(char,d==0)
static long th_char_p_store( void *addr, long d )	TH_type_STORE(char *,d==0)
static long th_u_short_store( void *addr, long d )	TH_type_STORE(unsigned short,0)
static long th_s_short_store( void *addr, long d )	TH_type_STORE(signed short,0)
static long th_u_int_store( void *addr, long d )	TH_type_STORE(unsigned int,0)
static long th_s_int_store( void *addr, long d )	TH_type_STORE(signed int,0)
static long th_u_long_store( void *addr, long d )	TH_type_STORE(unsigned long,0)
static long th_s_long_store( void *addr, long d )	TH_type_STORE(signed long,0)

#define TH_type_READ(type)	\
{\
	return( (long)(*((type *)addr)) );\
}
static long th_char_read( void *addr )		TH_type_READ(char)
static long th_char_p_read( void *addr )	TH_type_READ(char *)
static long th_u_short_read( void *addr )	TH_type_READ(unsigned short)
static long th_s_short_read( void *addr )	TH_type_READ(signed short)
static long th_u_int_read( void *addr )		TH_type_READ(unsigned int)
static long th_s_int_read( void *addr )		TH_type_READ(signed int)
static long th_u_long_read( void *addr )	TH_type_READ(unsigned long)
static long th_s_long_read( void *addr )	TH_type_READ(signed long)


struct typehandler_t th_s_char=
	{th_char_input,NULL,th_std_dup,
	 th_char_store,th_char_read,th_char_print};
struct typehandler_t th_u_char=
	{th_char_input,NULL,th_std_dup,
	 th_char_store,th_char_read,th_char_print};
struct typehandler_t th_s_short=
	{th_s_short_input,NULL,th_std_dup,
	 th_s_short_store,th_s_short_read,th_s_short_print};
struct typehandler_t th_u_short=
	{th_u_short_input,NULL,th_std_dup,
	 th_u_short_store,th_u_short_read,th_u_short_print};
struct typehandler_t th_s_int=
	{th_s_int_input,NULL,th_std_dup,
	 th_s_int_store,th_s_int_read,th_s_int_print};
struct typehandler_t th_u_int=
	{th_u_int_input,NULL,th_std_dup,
	 th_u_int_store,th_u_int_read,th_u_int_print};
struct typehandler_t th_s_long=
	{th_s_long_input,NULL,th_std_dup,
	 th_s_long_store,th_s_long_read,th_s_long_print};
struct typehandler_t th_u_long=
	{th_u_long_input,NULL,th_std_dup,
	 th_u_long_store,th_u_long_read,th_u_long_print};
struct typehandler_t th_s_char_p=
	{th_char_p_input,th_std_free,th_char_p_dup,
	 th_char_p_store,th_char_p_read,th_char_p_print};
struct typehandler_t th_u_char_p=
	{th_char_p_input,th_std_free,th_char_p_dup,
	 th_char_p_store,th_char_p_read,th_char_p_print};
struct typehandler_t th_void=
	{NULL,NULL,NULL,
	 NULL,NULL,NULL};
struct typehandler_t th_void_p=
	{NULL,th_std_free,NULL,
	 NULL,NULL,NULL};


/* name je deklaracia premennej aj s menom.
 * priklad: "unsigned int pole[3]"
 * 	    "char pole[]" == "char *pole"
 * 	    "char pole[3][4]"
 * 	    "int i"
 */
char *type_name( const char *name )
{ const char *p;
  int l,i,j,m,z;
  static char buf[64];
	p=name+strlen(name);
	m=0; z=0;
	while( p>name )
	{	if( z==0 && *p==']' )		z=1;
		else if( z==1 && *p=='[' )	z=2;
		else if( z==1 )			z=-1;
		if( m==0 && isalpha(*p) )	m=1;
		else if( m==1 && (*p==' ' || *p=='\t') )	m=2;
		else if( *p=='*' )
		{	p++;	break;	}
		else if( m==2 && !(*p==' ' || *p=='\t') )
		{	p++;	break;	}
		p--;
	}
	l=p-name;
	i=0; j=0; m=0;
	while( i<l && j<64-1 )
	{	if( m==0 && (name[i]==' ' || name[i]=='\t') )
		{	i++;	continue;	}
		if( name[i]==' ' || name[i]=='\t' )
		{	buf[j++]=' ';	i++;	m=0;	continue;	}
		buf[j++]=name[i++];
		m=1;
	}
	if( j>0 && j<64-2 && z==2 )
	{	if( buf[j-1]!=' ' && buf[j-1]!='*' )	buf[j++]=' ';
		buf[j++]='*';
	}
	buf[j]=0;
	return(buf);
}
char *type_varname( const char *name )
{ const char *p;
  int i,m;
  static char buf[64];
	p=name+strlen(name);
	m=0;
	while( p>name )
	{	if( m==0 && isalpha(*p) )	m=1;
		else if( m==1 && (*p==' ' || *p=='\t') )
		{	p++;	break;	}
		else if( *p=='*' )
		{	p++;	break;	}
		p--;
	}
	i=0;
	while( p[i]!=0 && p[i]!=' ' && p[i]!='\t' && p[i]!='[' && p[i]!=';' )
		i++;
	if( i>=64 )	i=64-1;
	buf[i]=0;
	i--;
	while( i>=0 )
	{	buf[i]=p[i];
		i--;
	}
	return(buf);
}
int *type_size( const char *name )
{ const char *p;
  int n;
  static int sizes[4];
	n=0;
	p=name;
	while( n<4-1 && (p=strchr(p,'['))!=NULL )
	{	p++;
		if( *p==']' )	continue;
		sizes[n++]=atoi(p);
	}
	if( n==0 )	return(NULL);
	sizes[n]=-1;
	return( sizes );
}




long type_input( int fd, const char *name, types_t *types )
{ int i;
  char *typ;
  int *sizes;
  int si[4];
  int j,sil,k;
  long size,pos,pd;
  long d;
  char *mem;
  char buf[64];
  	typ=type_name(name);
	i=0;
	while( types[i].name!=NULL && types[i].handler!=NULL )
	{	if( !declcmp(typ,types[i].name) )
		{	if( types[i].handler->input==NULL )
				return(0);
			if( (sizes=type_size(name))==NULL || sizes[0]<=0 )
			{	sprintf(buf,"%s=",type_varname(name));
				write(fd,buf,strlen(buf));
				return( types[i].handler->input(fd) );
			}
			size=1;
			for(sil=0;sil<4 && sizes[sil]>0;)
				size*=sizes[sil++];
			if( (mem=malloc(size))==NULL )
				return(0);
			for(j=0;j<sil;j++)	si[j]=0;
			pos=0;
			while( si[0]<sizes[0] )
			{	sprintf(buf,"%s",type_varname(name));
				write(fd,buf,strlen(buf));
				for(k=0;k<sil;k++)
				{	sprintf(buf,"[%d]",si[k]);
					write(fd,buf,strlen(buf));
				}
				sprintf(buf,"=");
				write(fd,buf,strlen(buf));
				d=types[i].handler->input(fd);
				pd=types[i].handler->store(mem+pos,d);
				if( pd > 0 )
				{	pos+=pd;	}
				else
				{	pos-=pd;	}
				j=sil-1;
   				si[j]++;
				while( j>0 && si[j]>=sizes[j] )
				{	si[j]=0;	j--;
					si[j]++;
				}
			}
			return((long)mem);
		}
		i++;
	}
printf("type %s not found!\n",typ);
	return(0);
}

int type_print( int fd, const char *name, types_t *types, long d )
{ int i;
  char *typ;
  int *sizes;
  int si[4];
  int j,sil,k;
  long pd;
  char buf[64];
  	typ=type_name(name);
//printf("type_print: d=%ld",d);
	i=0;
	while( types[i].name!=NULL && types[i].handler!=NULL )
	{	if( !declcmp(typ,types[i].name) )
		{	if( types[i].handler->print==NULL )
				return(-1);
			if( (sizes=type_size(name))==NULL || sizes[0]<=0 )
			{  snprintf(buf,64,"%s=",type_varname(name));
			   write(fd,buf,strlen(buf));
			   types[i].handler->print(fd,d);
			   return(0);
			}
			if( d==0 )
			{  snprintf(buf,64,"%s=(null)",type_varname(name));
			   write(fd,buf,strlen(buf));
			   return(0);
			}
			for(sil=0;sil<4 && sizes[sil]>0;)	sil++;
			for(j=0;j<sil;j++)	si[j]=0;
			while( si[0]<sizes[0] )
			{	sprintf(buf,"%s",type_varname(name));
				write(fd,buf,strlen(buf));
				for(k=0;k<sil;k++)
				{	sprintf(buf,"[%d]",si[k]);
					write(fd,buf,strlen(buf));
				}
				sprintf(buf,"=");
				write(fd,buf,strlen(buf));
				pd=types[i].handler->print(fd,
				 	types[i].handler->read((void*)d));
				write(fd," ",1);
				if( pd > 0 )
				{	d+=pd;	}
				else
				{	d-=pd;	}
				j=sil-1;
   				si[j]++;
				while( j>0 && si[j]>=sizes[j] )
				{	si[j]=0;	j--;
					si[j]++;
				}
			}
			return(0);
		}
		i++;
	}
	return(-1);
}

int type_free( const char *name, types_t *types, long d )
{ int i;
  char *typ;
  int *sizes;
  int si[4];
  long dd,pd;
  int j,sil;
  	dd=d;
  	typ=type_name(name);
	i=0;
	while( types[i].name!=NULL && types[i].handler!=NULL )
	{	if( !declcmp(typ,types[i].name) )
		{	if( types[i].handler->free==NULL )
				return(0);
			if( (sizes=type_size(name))==NULL || sizes[0]<=0 )
			{  types[i].handler->free(d);
			   return(0);
			}
			if( d==0 )	return(0);
			for(sil=0;sil<4 && sizes[sil]>0;)	sil++;
			for(j=0;j<sil;j++)	si[j]=0;
			while( si[0]<sizes[0] )
			{	pd=types[i].handler->store(NULL,
				 	types[i].handler->read((void*)d));
				types[i].handler->free(
					types[i].handler->read((void*)d) );
				if( pd > 0 )
				{	d+=pd;	}
				else
				{	d-=pd;	}
				j=sil-1;
   				si[j]++;
				while( j>0 && si[j]>=sizes[j] )
				{	si[j]=0;	j--;
					si[j]++;
				}
			}
			free((void*)dd);
			return(0);
		}
		i++;
	}
	return(-1);
}

long type_dup( const char *name, types_t *types, long d )
{ int i;
  char *typ;
  int *sizes;
  int si[4];
  char *mem;
  int size,pos;
  long dd,pd,pd2;
  int j,sil;
  	dd=d;
  	typ=type_name(name);
	i=0;
	while( types[i].name!=NULL && types[i].handler!=NULL )
	{	if( !declcmp(typ,types[i].name) )
		{	if( types[i].handler->dup==NULL )
				return(0);
			if( (sizes=type_size(name))==NULL || sizes[0]<=0 )
			{  return( types[i].handler->dup(d) );
			}
			if( d==0 )	return(0);
			size=1;
			for(sil=0;sil<4 && sizes[sil]>0;)
				size*=sizes[sil++];
			if( (mem=malloc(size))==NULL )
				return(0);
			pos=0;
			for(j=0;j<sil;j++)	si[j]=0;
			while( si[0]<sizes[0] )
			{	pd=types[i].handler->store(NULL,
				 	types[i].handler->read((void*)d));
				pd2=types[i].handler->store(mem,
					types[i].handler->dup(
					   types[i].handler->read((void*)d) ));
				if( pd > 0 )
				{	d+=pd;	}
				else
				{	d-=pd;	}
				if( pd2 > 0 )
				{	pos+=pd;	}
				else
				{	pos-=pd;	}
				j=sil-1;
   				si[j]++;
				while( j>0 && si[j]>=sizes[j] )
				{	si[j]=0;	j--;
					si[j]++;
				}
			}
			return((long)mem);
		}
		i++;
	}
	return(0);
}

long type_str_input( char *buf, const char *name, types_t *types )
{ int fd[2];
  long r;
	if( pipe(fd)<0 )	return(0);
	write(fd[1],buf,strlen(buf));
	close(fd[1]);
	r=type_input(fd[0],name,types);
	close(fd[0]);
	return(r);
}

/*
int type_print_struct( int fd, struct_t *desc, types_t *types, void *s )
{ int i,
  int ofs;
	ofs=0;
	i=0;
	while( desc[i].type!=NULL )
	{	

	}
	return(0);
}
*/
