/* User Interface - types
 * (c) by Marek Zelem
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <mmm/menu.h>

void th_std_free( long d, int size )
{
	if( d==0 || size<0 )	return;
	free((char*)d);
}

long th_char_input( int fd, int size )
{ char c;
  char *p=NULL;
  int pos=0;
  	if( size>0 )
	{	if( (p=malloc(size))==NULL )
			return(0);
	}
  	while( read(fd,&c,1)==1 )
	{	if( size==TYPESIZE_VAR )
			return((long)c);
		if( size>0 )
		{	p[pos++]=c;	size--;
			if( size==0 )	return((long)p);
		}
		if( size==0 )
		{	if( p==NULL )
			{	if( (p=malloc(pos+2))==NULL )
					return(0);
			}
			else
			{	if( (p=realloc(p,pos+2))==NULL )
					return(0);
			}
			p[pos++]=c;
		}
		if( c==0 || c=='\n' )
		{	if( c=='\n' && pos>0 )	pos--;
			p[pos]=0;
			return((long)p);
		}
	}
	if( p!=NULL )
		free(p);
	return(0);
}
void th_char_print( int fd, long d, int size )
{ char c;
  int i;
	if( size==TYPESIZE_VAR )
	{	c=(char)d;
		write(fd,&c,1);
	}
	if( size==0 && d!=0 )
	{	write(fd,(char*)d,strlen((char*)d));
	}
	if( size>0 && d!=0 )
	{	for(i=0;i<size && ((char*)d)[i]!=0;)	i++;
		write(fd,(char*)d,i);
	}
	if( size>=0 && d==0 )
	{	write(fd,"(null)",6);
	}
}

static char *read_string( int fd )
{ char *p,*pp;
  int len,size;
  	len=0;
	size=16;
	if( (p=malloc(size))==NULL )	return(NULL);
  	while( read(fd,p+len,1)==1 )
	{	if( p[len]==0 || p[len]=='\n' )
			break;
		len++;
		if( len>=size )
		{	size=len+16;
			if( (p=realloc((pp=p),size))==NULL )
			{	pp[len-1]=0;
				return(pp);
			}
		}
	}
	p[len]=0;
	return(p);
}

long th_char_p_input( int fd, int size )
{ char *c;
  char **p=NULL;
  int pos=0;
  	if( size>0 )
	{	if( (p=malloc(size*sizeof(char*)))==NULL )
			return(0);
	}
  	while( (c=read_string(fd))!=NULL )
	{	if( size==TYPESIZE_VAR )
			return((long)c);
		if( size>0 )
		{	p[pos++]=c;	size--;
			if( size==0 )	return((long)p);
		}
		if( size==0 )
		{	if( p==NULL )
			{	if( (p=malloc((pos+2)*sizeof(char*)))==NULL )
					return(0);
			}
			else
			{	if( (p=realloc(p,(pos+2)*sizeof(char*)))==NULL )
					return(0);
			}
			p[pos++]=c;
		}
		if( c==NULL || c[0]==0 )
		{	if( pos>0 )	pos--;
			p[pos]=NULL;
			return((long)p);
		}
	}
	if( p!=NULL )
	{	pos--;
		while( pos>=0 )
			free(p[pos--]);
		free(p);
	}
	return(0);
}
void th_char_p_free( long d, int size )
{ char **p;
	if( d==0 )	return;
	if( size==TYPESIZE_VAR )
	{	free((char*)d);
		return;
	}
	p=(char**)d;
	if( size>=0 )
	{	while( *p!=NULL )
		{	free(*p);
			p++;
			if( size>0 )
			{	size--;
				if( size==0 )
					break;
			}
		}
		free((char*)d);
	}
}


static int conv_zn( char c )
{
	if( c>='0' && c<='9' )
		return(c-'0');
	if( c>='a' && c<='f' )
		return(c-'a'+10);
	return(1000);
}

static long read_long( int fd )
{ long d=0,i;
  int base=10,sign=1,z=0;
  char c;
  	while( read(fd,&c,1)==1 )
	{	if( z==0 && c=='-' )	sign=-1;
		else if( z==0 && c=='0' )
		{	base=8;	z=1;	}
		else if( (z==0 || z==1) && c=='x' )
		{	base=16; z=2;	}
		else if( (i=(long)conv_zn(tolower(c)))<base )
		{	d*=base;
			d+=i;
			z=2;
		}
		else	return(d*sign);
	}
	return(d*sign);
}

#define TH_type_INPUT(type)	\
{ type c;\
  type *p=NULL;\
  int pos=0;\
  	if( size>0 )\
	{	if( (p=malloc(size*sizeof(type)))==NULL )\
			return(0);\
	}\
	if( size==TYPESIZE_VAR )\
	{	c=(type)(read_long(fd));\
		return((long)c);\
	}\
  	while( size>0 )\
	{	c=(type)(read_long(fd));\
		p[pos++]=c;	size--;\
	}\
	return((long)p);\
}
long th_u_short_input( int fd, int size )	TH_type_INPUT(unsigned short)
long th_s_short_input( int fd, int size )	TH_type_INPUT(signed short)
long th_u_int_input( int fd, int size )		TH_type_INPUT(unsigned int)
long th_s_int_input( int fd, int size )		TH_type_INPUT(signed int)
long th_u_long_input( int fd, int size )	TH_type_INPUT(unsigned long)
long th_s_long_input( int fd, int size )	TH_type_INPUT(signed long)

#define TH_type_PRINT(pr,type)	\
{ int i;\
  char buf[64];\
	if( size==TYPESIZE_VAR )\
	{	sprintf(buf,"%"pr,(type)d);\
		write(fd,buf,strlen(buf));\
	}\
	if( size>0 && d!=0 )\
	{	for(i=0;i<size;i++)\
		{	sprintf(buf,"%"pr" ",(((type *)d)[i]));\
			write(fd,buf,strlen(buf));\
		}\
	}\
	if( size>=0 && d==0 )\
	{	write(fd,"(null)",6);\
	}\
}
void th_u_short_print( int fd, long d, int size )
	TH_type_PRINT("u",unsigned short)
void th_s_short_print( int fd, long d, int size )
	TH_type_PRINT("d",signed short)
void th_u_int_print( int fd, long d, int size )
	TH_type_PRINT("u",unsigned int)
void th_s_int_print( int fd, long d, int size )
	TH_type_PRINT("d",signed int)
void th_u_long_print( int fd, long d, int size )
	TH_type_PRINT("lu",unsigned long)
void th_s_long_print( int fd, long d, int size )
	TH_type_PRINT("ld",signed long)
void th_char_p_print( int fd, long d, int size )
	TH_type_PRINT("s",char *)


struct typehandler_t th_s_char=
		{th_char_input,th_char_print,th_std_free};
struct typehandler_t th_u_char=
		{th_char_input,th_char_print,th_std_free};
struct typehandler_t th_s_short=
		{th_s_short_input,th_s_short_print,th_std_free};
struct typehandler_t th_u_short=
		{th_u_short_input,th_u_short_print,th_std_free};
struct typehandler_t th_s_int=
		{th_s_int_input,th_s_int_print,th_std_free};
struct typehandler_t th_u_int=
		{th_u_int_input,th_u_int_print,th_std_free};
struct typehandler_t th_s_long=
		{th_s_long_input,th_s_long_print,th_std_free};
struct typehandler_t th_u_long=
		{th_u_long_input,th_u_long_print,th_std_free};
struct typehandler_t th_s_char_p=
		{th_char_p_input,th_char_p_print,th_char_p_free};
struct typehandler_t th_u_char_p=
		{th_char_p_input,th_char_p_print,th_char_p_free};


/* name je deklaracia premennej aj s menom.
 * priklad: "unsigned int pole[3]"
 * 	    "char pole[]"
 * 	    "char *pole"
 * 	    "int i"
 */
char *type_name( const char *name )
{ const char *p;
  int l,i,j,m;
  static char buf[64];
	p=name+strlen(name);
	m=0;
	while( p>name )
	{	if( m==0 && isalpha(*p) )	m=1;
		else if( m==1 && (*p==' ' || *p=='\t') )	m=2;
		else if( *p=='*' )
		{	p++;	break;	}
		else if( m==2 && !(*p==' ' || *p=='\t') )
		{	p++;	break;	}
		p--;
	}
	l=p-name;
	i=0; j=0; m=0;
	while( i<l && j<64 )
	{	if( m==0 && (name[i]==' ' || name[i]=='\t') )
		{	i++;	continue;	}
		if( name[i]==' ' || name[i]=='\t' )
		{	buf[j++]=' ';	i++;	m=0;	continue;	}
		buf[j++]=name[i++];
		m=1;
	}
	buf[j]=0;
	return(buf);
}
char *type_varname( const char *name )
{ const char *p;
  int i,m;
  static char buf[64];
	p=name+strlen(name);
	m=0;
	while( p>name )
	{	if( m==0 && isalpha(*p) )	m=1;
		else if( m==1 && (*p==' ' || *p=='\t') )
		{	p++;	break;	}
		else if( *p=='*' )
		{	p++;	break;	}
		p--;
	}
	i=0;
	while( p[i]!=0 && p[i]!=' ' && p[i]!='\t' && p[i]!='[' && p[i]!=';' )
		i++;
	if( i>=64 )	i=64-1;
	buf[i]=0;
	i--;
	while( i>=0 )
	{	buf[i]=p[i];
		i--;
	}
	return(buf);
}
int type_size( const char *name )
{ const char *p;
	if( (p=strchr(name,'['))!=NULL )
	{	p++;
		return(atoi(p));
	}
	return( TYPESIZE_VAR );
}




long type_input( int fd, const char *name, types_t *types )
{ int i;
  char *typ;
  char buf[64];
  	typ=type_name(name);
	i=0;
	while( types[i].name!=NULL && types[i].handler!=NULL )
	{	if( !strcmp(typ,types[i].name) )
		{	if( types[i].handler->input==NULL )
				return(0);
			sprintf(buf,"%s=",type_varname(name));
			write(fd,buf,strlen(buf));
			return( types[i].handler->input(fd,type_size(name)) );
		}
		i++;
	}
printf("type %s not found!\n",typ);
	return(0);
}

int type_print( int fd, const char *name, types_t *types, long d )
{ int i;
  char *typ;
  char buf[64];
  	typ=type_name(name);
printf("type_print: d=%ld size=%d\n",d,type_size(name));
	i=0;
	while( types[i].name!=NULL && types[i].handler!=NULL )
	{	if( !strcmp(typ,types[i].name) )
		{	if( types[i].handler->print==NULL )
				return(-1);
			snprintf(buf,64,"%s=",type_varname(name));
			write(fd,buf,strlen(buf));
			types[i].handler->print(fd,d,type_size(name));
			return(0);
		}
		i++;
	}
	return(-1);
}

int type_free( const char *name, types_t *types, long d )
{ int i;
  char *typ;
  	typ=type_name(name);
	i=0;
	while( types[i].name!=NULL && types[i].handler!=NULL )
	{	if( !strcmp(typ,types[i].name) )
		{	if( types[i].handler->free==NULL )
				return(0);
			types[i].handler->free(d,type_size(name));
			return(0);
		}
		i++;
	}
	return(-1);
}
/*
int type_print_struct( int fd, struct_t *desc, types_t *types, void *s )
{ int i,
  int ofs;
	ofs=0;
	i=0;
	while( desc[i].type!=NULL )
	{	

	}
	return(0);
}
*/
