/* argument parser	*/
/* (c) by Marek Zelem	*/

#include <mmm/text.h>
#include <mmm/text_lib.h>
#include <mmm/logic.h>
#include <mmm/dynamic.h>
#include <string.h>
#include <stdlib.h>

typedef struct	{	
			int from,frome;	/* predchadzajuci stav(nie celkom) */
			int vp;		/* pozicia vo vzore */
			int a,ap;	/* cislo argumentu a pozicia v nom */
			int app;	/* predch. pozicia v arg. */
			int n;		/* poradove cislo retazca ktory sadol*/
			int repeat;	/* ci opakovat aktualnu uroven */
			int soutn,ioutn;/* stav out na zaciatku vnorenia */
			int ignore;	/* ignorovat nezhody
					   % nastavit na NULL */
		} _argpp_t;

typedef struct	{
			char *vzor;	/* format argumentov */
			char **arg;	/* vstupne argumenty */
			_argpp_t p;
			dfifo_t *stack;
			darg_t *sout;	/* vystupne retazce */
			ddata_t *iout;	/* vystupne cisla */
		} _argpd_t;

int _vyraz( void *data );
int _pop( void *data );
int _poznamka( void *data );
int _retazec( void *data );
int _mretazec( void *data );
int _mretazec_next( void *data );
int _mretazec_noeql( void *data );
int _mretazec_finish( void *data );
int _input( void *data );
int _input_word( void *data );
int _input_num( void *data );

void _argpd_datafree( void *data )
{ _argpd_t *apd=data;
	if( apd!=NULL && apd->stack!=NULL )
		dfifo_delete(apd->stack);
	if( apd!=NULL )	free(apd);
}

enum	{	END,		/* lebo 0 */
		VYRAZ,
		POZNAMKA,
		RETAZEC,
		MRETAZEC, MRETAZEC_FINISH, MRETAZEC_NEXT, MRETAZEC_NOEQL, 
		INPUT, INPUT_WORD, INPUT_NUM,
		POP
	};

automat_status_t arg_parser_states[]=	{
		{VYRAZ,_vyraz},
		{POZNAMKA,_poznamka},
		{RETAZEC,_retazec},
		{MRETAZEC,_mretazec},
		{MRETAZEC_FINISH,_mretazec_finish},
		{MRETAZEC_NEXT,_mretazec_next},
		{MRETAZEC_NOEQL,_mretazec_noeql},
		{INPUT,_input},
		{INPUT_WORD,_input_word},
		{INPUT_NUM,_input_num},
		{POP,_pop},
		{-1,NULL}
					};

int arg_parse( char *format, char *argv[], darg_t **sout, ddata_t **iout )
{ _argpd_t *data;
  automat_t *au;
  int r;
//printf("DEBUG: arg_parse: format='%s'\n",format);
//printf("DEBUG: arg_parse: argv[0]='%s'  argv[1]='%s'\n",argv[0],argv[1]);
	*sout=NULL; *iout=NULL;
	if( (data=malloc(sizeof(_argpd_t)))==NULL )
	{	return(-1);	}
	data->vzor=format;
	data->arg=argv;
	data->p.vp=0;
	data->p.a=0;
	data->p.ap=0;
	data->p.repeat=0;
	data->p.ignore=0;
	data->p.from= -1;
	data->p.frome= -1;
	if( (data->stack=dfifo_create(sizeof(_argpp_t),8))==NULL )
	{	free( data );	return(-1);	}
	if( (data->sout=darg_alloc(1))==NULL )
	{	dfifo_delete(data->stack); free(data); return(-1);	}
	data->p.soutn=darg_argc(data->sout);
	if( (data->iout=dd_alloc(sizeof(int),16))==NULL )
	{	darg_free(data->sout);
		dfifo_delete(data->stack); free(data); return(-1);	}
	data->p.ioutn=dd_n(data->iout);
	if( (au=automat_create(arg_parser_states))==NULL )
	{	dd_free(data->iout); darg_free(data->sout);
		dfifo_delete(data->stack); free(data); return(-1);	}
	automat_set_data(au,data,_argpd_datafree);
	r=automat_start(au,VYRAZ);
	automat_destroy(au);
	*iout=data->iout; *sout=data->sout;
//printf("DEBUG arg_parse return %d\n",r);
	return(r==0?1:(r==-2?0:r));
}

int arg_parse_free( darg_t **sout, ddata_t **iout )
{
	if( iout!=NULL && *iout!=NULL )
		dd_free(*iout);
	if( sout!=NULL && *sout!=NULL )
		darg_free(*sout);
	return(0);
}

int _vyraz( void *data )
{ _argpd_t *apd=data;
  char c;
  _argpp_t appnull;
//printf("_vyraz %d\n",dfifo_n(apd->stack));
//printf("       vzor='%s'\n",&(apd->vzor[apd->p.vp]));
//if( apd->arg[apd->p.a]!=NULL )
//printf("       %d arg='%s'\n",apd->p.ap,&(apd->arg[apd->p.a][apd->p.ap]));
	apd->p.from=VYRAZ;
	apd->p.frome=VYRAZ;
	if( (c=apd->vzor[(apd->p.vp)++])==0 && dfifo_n(apd->stack)==0 )
			return(0);
	if( c=='$' && apd->p.ignore==0 )
	{	if( apd->arg[apd->p.a]==NULL )
			return(VYRAZ);
		if( apd->arg[apd->p.a][apd->p.ap]==0 
		    && apd->arg[apd->p.a+1]==NULL )
		{	apd->p.ap=0;
			apd->p.a++;
			return(VYRAZ);
		}
		return(POP);
	}
	if( c=='<' )	return(POZNAMKA);
	if( c==' ' && apd->p.ignore==0 )
	{	if( apd->arg[apd->p.a][apd->p.ap]==0 )
		{	apd->p.ap=0;
			apd->p.a++;
			return(VYRAZ);
		}
		else	return(POP);
	}
	if( c=='[' )
	{	if( apd->vzor[apd->p.vp]==':' )
		{	(apd->p.vp)++;
			dfifo_write(apd->stack,&(apd->p),1);
			apd->p.soutn=darg_argc(apd->sout);
			apd->p.ioutn=dd_n(apd->iout);
			apd->p.repeat=1;
			return(VYRAZ);
		}
		else
		{	dfifo_write(apd->stack,&(apd->p),1);
			apd->p.soutn=darg_argc(apd->sout);
			apd->p.ioutn=dd_n(apd->iout);
			apd->p.repeat=0;
			return(VYRAZ);
		}
	}
	if( c==']' && apd->p.repeat==0 && dfifo_n(apd->stack)>0 )
	{	dlifo_read(apd->stack,&appnull,1);
		apd->p.repeat=appnull.repeat;
		apd->p.ignore=appnull.ignore;
		apd->p.soutn=appnull.soutn;
		apd->p.ioutn=appnull.ioutn;
		return(VYRAZ);
	}
	if( c==':' && apd->p.repeat && apd->vzor[apd->p.vp]==']' )
	{	(apd->p.vp)++;
		if( apd->p.ignore==0 )
		{	dlifo_read(apd->stack,&appnull,1);
			apd->p.vp=appnull.vp;
			appnull.a=apd->p.a;
			appnull.ap=apd->p.ap;
			dfifo_write(apd->stack,&appnull,1);
			apd->p.soutn=darg_argc(apd->sout);
			apd->p.ioutn=dd_n(apd->iout);
		}
		else
		{	dlifo_read(apd->stack,&appnull,1);
			apd->p.repeat=appnull.repeat;
			apd->p.ignore=appnull.ignore;
		}
		return(VYRAZ);
	}
	if( apd->arg[apd->p.a]==NULL && apd->p.ignore==0 )	return(POP);
	if( c=='"' )	return(RETAZEC);
	if( c=='{' )	return(MRETAZEC);
	if( c=='%' )	return(INPUT);
	if( apd->p.ignore )	return(VYRAZ);
	return(-1);	/* syntax error */
}

int _pop( void *data )
{ _argpd_t *apd=data;
  int n,repeat,frome;
//printf("_pop %d\n",dfifo_n(apd->stack));
	if( dfifo_n(apd->stack)<=0 )	return(-2);	/* No match */
	repeat=apd->p.repeat;
	frome=apd->p.frome;
	if( frome==VYRAZ )
	{	while( (n=darg_argc(apd->sout)-apd->p.soutn)>0 )
			darg_delete(apd->sout,apd->p.soutn);
		if( (n=dd_n(apd->iout)-apd->p.ioutn)>0 )
			dd_delete(apd->iout,apd->p.ioutn,n);
		dfifo_last(apd->stack,&(apd->p),1);
		apd->p.repeat=repeat;
		apd->p.ignore=1;
		return(VYRAZ);
	}
	dlifo_read(apd->stack,&(apd->p),1);
	return( frome );	/* ??? !!!! DOMYSLIET, a mozno ani nie */
}

int _poznamka( void *data )
{ _argpd_t *apd=data;
  char c;
	while( (c=apd->vzor[(apd->p.vp)++])!=0 && c!='>' );
	if( c=='>' )	return(apd->p.from);
	return(-1);	/* syntax error */
}

int _retazec( void *data )
{ _argpd_t *apd=data;
  char c;
  long l;
//printf("_retazec %d\n",dfifo_n(apd->stack));
	while( (c=apd->vzor[(apd->p.vp)++])!=0 && c!='"' )
	{	l=0;
		if( c=='\\' )
		{	c=tex_parse_backslash(&(apd->vzor[(apd->p.vp)-1]),&l);
			apd->p.vp+=l-1;	l=1;
		}
		if( apd->p.ignore )	continue;
		if( l==0 && c=='[' )
		{ char_fd_t cf;
		    apd->p.vp+=tex_charfd_init(&(apd->vzor[apd->p.vp]),&cf);
		    if( apd->vzor[apd->p.vp]==']' )	apd->p.vp+=1;
		    else	return(-1);	/* syntax error */
		    if( tex_charfd_isset(apd->arg[apd->p.a][apd->p.ap],&cf) )
		    {	apd->p.ap++;	continue;	}
		    return(POP);
		}
/*	!!! DOKONCIT + {} + *
			... mozno skok do dalsieho stavu a zapamatanie navratu
			alebo na to cele hodit wildcmp ...?
*/
		if( (l==0 && c=='?' && apd->arg[apd->p.a][apd->p.ap]!=0 )
			|| apd->arg[apd->p.a][apd->p.ap]==c )
		{	apd->p.ap++;	continue;	}
		return(POP);	/* not match */
	}
	if( c=='"' )	return(apd->p.from);
	return(-1);	/* syntax error */
}

int _mretazec( void *data )
{ _argpd_t *apd=data;
  char c;
//printf("_mretazec %d\n",dfifo_n(apd->stack));
	if( (c=apd->vzor[(apd->p.vp)++])=='}' )
	{	return(apd->p.from);
	}
	if( c=='"' )
	{	dfifo_write(apd->stack,&(apd->p),1);
		apd->p.from=MRETAZEC_FINISH;
		apd->p.frome=MRETAZEC_NOEQL;
		return(RETAZEC);
	}
	return(-1);	/* syntax error */
}

int _mretazec_next( void *data )
{ _argpd_t *apd=data;
  char c;
  _argpp_t appnull;
//printf("_mretazec_next %d\n",dfifo_n(apd->stack));
	dlifo_read(apd->stack,&(appnull),1);
  	apd->p.ignore=appnull.ignore;	/* je to vzdy 0 */
	apd->p.from=appnull.from;
	apd->p.frome=appnull.frome;
	if( (c=apd->vzor[(apd->p.vp)++])=='}' )
		return(POP);	/* no match */
	if( c==',' )
	{	apd->p.n+=1;
		return(MRETAZEC);
	}
	if( c=='|' )
		return(MRETAZEC);
	return(-1);	/* syntax error */
}

int _mretazec_noeql( void *data )
{ _argpd_t *apd=data;
//printf("_mretazec_noeql %d\n",dfifo_n(apd->stack));
	/* ignore je teraz urc~ite 0 , lebo inac~ by sa sem nedostal */
	dfifo_write(apd->stack,&(apd->p),1);
	apd->p.ignore=1;
	apd->p.from=MRETAZEC_NEXT;
	return(RETAZEC);
}

int _mretazec_finish( void *data )
{ _argpd_t *apd=data;
  char c;
  _argpp_t appnull;
//printf("_mretazec_finish %d\n",dfifo_n(apd->stack));
	if( (c=apd->vzor[(apd->p.vp)++])=='}' )
	{	dlifo_read(apd->stack,&(appnull),1);
		apd->p.ignore=appnull.ignore;
		apd->p.from=appnull.from;
		apd->p.frome=appnull.frome;
		return(apd->p.from);
	}
	if( c==',' || c=='|' )
	{
		if( (c=apd->vzor[(apd->p.vp)++])=='}' )
		{	dlifo_read(apd->stack,&(appnull),1);
			apd->p.ignore=appnull.ignore;
			apd->p.from=appnull.from;
			apd->p.frome=appnull.frome;
			return(apd->p.from);
		}
		if( c=='"' )
		{
			apd->p.ignore=1;
			apd->p.from=MRETAZEC_FINISH;
			return(RETAZEC);
		}
		return(-1);	/* syntax error */
	}
	return(-1);	/* syntax error */
}

int _input_oneof( char **data, int *m, char *in, int *len)
{ int i,l,j;
	for(j=0;m[j]>0;j++)
	{	for(i=0;data[i]!=NULL;i++)
		{	l=strlen(data[i]);
			if( m[j]<l )	l=m[j];
			if( !strncasecmp(data[i],in,l) )
			{	if( len!=NULL )	*len=l;
				return(i);
			}
		}
	}
	if( len!=NULL )	*len=0;
	return(-1);
}

int _input( void *data )
{ _argpd_t *apd=data;
  char c;
  int i,n;
  char_fd_t char_fd;
//printf("_input %d\n",dfifo_n(apd->stack));
	if( (c=apd->vzor[(apd->p.vp)++])=='{' )
	{	apd->p.n=0;
		apd->p.from=INPUT_NUM;
		return(MRETAZEC);
	}
	if( c=='"' )
	{	apd->p.app=apd->p.ap;
		apd->p.from=INPUT_WORD;
		return(RETAZEC);
	}
	if( c=='d' )
	{	n=-1;
		if( !apd->p.ignore )
		{	n=0; i=0;
			while( (c=apd->arg[apd->p.a][apd->p.ap])!=0 )
			{	apd->p.ap++;
				if( c<'0' || c>'9' )
				{	apd->p.ap-=1;
					break;
				}
				n*=10;	i++;
				n+=c-'0';
			}
			if( i==0 )	return(POP);
		}
		dd_append(apd->iout,&n,1);
		return(VYRAZ);
	}
	if( c=='D' )
	{	char *data[]={	"Sunday","Monday","Tuesday","Wednesday",
				"Thursday","Friday","Saturday",NULL};
		int dl[]={100,3,2,0};
		n=-1;
		if( !apd->p.ignore )
		{
		n=_input_oneof(data,dl,&(apd->arg[apd->p.a][apd->p.ap]),&i);
		if( i==0 )	return(POP);
		apd->p.ap+=i;
		}
		dd_append(apd->iout,&n,1);
		return(VYRAZ);
	}
	if( c=='M' )
	{	char *data[]={	"January","February","March","April","May",
				"June","July","August","September","October",
				"November","December",NULL };
		int dl[]={100,3,0};
		n=-1;
		if( !apd->p.ignore )
		{
		n=_input_oneof(data,dl,&(apd->arg[apd->p.a][apd->p.ap]),&i);
		if( i==0 )	return(POP);
		apd->p.ap+=i;
		}
		dd_append(apd->iout,&n,1);
		return(VYRAZ);
	}
	if( c=='w' )
	{	tex_charfd_init("a-zA-Z",&char_fd);
		goto Get_word;
	}
	if( c=='[' )
	{	apd->p.vp+=tex_charfd_init(&(apd->vzor[apd->p.vp]),&char_fd);
		if( apd->vzor[apd->p.vp]==']' )	apd->p.vp+=1;
		else	return(-1);	/* syntax error */
Get_word:	if( !apd->p.ignore )
		{	apd->p.app=apd->p.ap;
			while( (c=apd->arg[apd->p.a][apd->p.ap])!=0 )
			{	if( !tex_charfd_isset(c,&char_fd) )
					break;
				apd->p.ap+=1;
			}
			if( apd->p.app==apd->p.ap )	return(POP);
		}
		return(INPUT_WORD);
	}
	return(-1);	/* syntax error */
}

int _input_word( void *data )
{ _argpd_t *apd=data;
  char c;
	if( apd->p.ignore )
	{	darg_append(apd->sout,NULL);
		return(VYRAZ);
	}
  	c=apd->arg[apd->p.a][apd->p.ap];
  	apd->arg[apd->p.a][apd->p.ap]=0;	/* !!! POZOR prepisujem */
	darg_append(apd->sout,&(apd->arg[apd->p.a][apd->p.app]));
  	apd->arg[apd->p.a][apd->p.ap]=c;
	return(VYRAZ);
}

int _input_num( void *data )
{ _argpd_t *apd=data;
  int n;
	if( apd->p.ignore )	n=-1;
	else			n=apd->p.n;
	dd_append(apd->iout,&n,1);
	return(VYRAZ);
}

